#include <windows.h>
#include <math.h> //Floating point operations

//Default Parameters
WORD Width=600, Height=400; //Window Dimensions
double XMin=-2, XMax=1, YMin=-1, YMax=1; //The X/Y axis min/max points
int MaxIterations=100; //Detail Level

//Program specific global variables
HDC BackBuffer;

//Function headers
void CreateBackBuffer(HWND hWnd); //Creates a backbuffer of needed size for the window
void DrawWindow(HWND hWnd); //Draws the fractal into the backbuffer and outputs it to the window
void DrawFractal(DWORD *PixelArray); //Does the actual fractal drawing into a pixel array
DWORD ProcessColor(int IterationsToEscape, double ValueAtEndReal, double ValueAtEndImaginary, double BailoutRadius, double Power); //Determine the color of a pixel based on its state
LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam); //Process windows events

int APIENTRY WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR pCmdLine,int nCmdShow)
{
	//Create the window class
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), NULL, WndProc, 0L, 0L, hInst, LoadIcon(hInst, "MAIN_ICON"), LoadCursor(NULL,IDC_ARROW), NULL, NULL, "JuliaFractal", NULL };
	if(RegisterClassEx(&wc)==0)
		return false;

	//Determine the size of the actual window that we want based upon client area dimensions
	RECT WindowRect={0, 0, Width, Height};
	AdjustWindowRect(&WindowRect, WS_OVERLAPPEDWINDOW, false);

	//Create the window and display it
	HWND MainHwnd = CreateWindowEx(NULL, "JuliaFractal", "Julia Fractal", WS_OVERLAPPEDWINDOW, 0, 0, WindowRect.right-WindowRect.left, WindowRect.bottom-WindowRect.top, NULL, NULL, hInst, NULL);
	ShowWindow(MainHwnd, SW_SHOW);
	UpdateWindow(MainHwnd);

	//Create the backbuffer device
	HDC MainDC=GetDC(MainHwnd);
	BackBuffer=CreateCompatibleDC(MainDC);
	ReleaseDC(MainHwnd, MainDC);
	CreateBackBuffer(MainHwnd);

	//Draw the main window for the first time
	DrawWindow(MainHwnd);

	//Main loop listening for events
	MSG msg;
	ZeroMemory(&msg, sizeof(MSG));
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg); 
		DispatchMessage(&msg);
	}

	//Destroy windows handles
	DeleteDC(BackBuffer);
	UnregisterClass("JuliaFractal", hInst);

	return S_OK;
}

void CreateBackBuffer(HWND hWnd) //Creates a backbuffer of needed size for the window
{
	HDC MainDC=GetDC(hWnd);
	HBITMAP NewBitmap=CreateCompatibleBitmap(MainDC, Width, Height);
	SelectObject(BackBuffer, NewBitmap);
	DeleteObject(NewBitmap);
	ReleaseDC(hWnd, MainDC);
}

void DrawWindow(HWND hWnd) //Draws the fractal into the backbuffer and outputs it to the window
{
	//Create the Fractal
	DWORD *FractalArray=new DWORD[Width*Height]; //Create the pixel array
	DrawFractal(FractalArray); //Do the drawing

	//Output to BackBuffer
	BITMAPINFO BitmapInfo={sizeof(BITMAPINFOHEADER), Width, Height, 1, 32, BI_RGB, 0, 0, 0, 0, 0};
	SetDIBits(BackBuffer, (HBITMAP)GetCurrentObject(BackBuffer, OBJ_BITMAP), 0, Height, FractalArray, &BitmapInfo, DIB_RGB_COLORS);

	//Clear out pixel array
	delete[] FractalArray;

	//Draw the backbuffer to the main screen
	HDC TheDC=GetDC(hWnd);
	BitBlt(TheDC,0,0,Width,Height,BackBuffer,0,0,SRCCOPY);
	ReleaseDC(hWnd, TheDC);
}

void DrawFractal(DWORD *PixelArray) //Does the actual fractal drawing into a pixel array
{
	for(int Y=0;Y<Height;Y++)
		for(int X=0;X<Width;X++)
		{
			//Determine point in the 2D fractal plane we are checking and assign to {x,y} (used for break out testing) and {x0,y0} (which remember the original coordinate)
			double x0=(double)(XMax-XMin)*((double)X/Width)+XMin, y0=(YMax-YMin)*((double)Y/Height)+YMin, x=x0, y=y0;

			//Determine if and when the point breaks out
			for(int iteration=0; x*x+y*y<=(2*2) && iteration<MaxIterations; iteration++)
			{
				double xtemp=x*x-y*y+x0;
				y=2*x*y+y0;
				x=xtemp;
			}

			//Assign the pixel color depending on if and when the point breaks out
			PixelArray[Y*Width+X]=(iteration == MaxIterations ? 0 : ProcessColor(iteration, x, y, 2, 2));
		}
}

DWORD ProcessColor(int IterationsToEscape, double ValueAtEndReal, double ValueAtEndImaginary, double BailoutRadius, double Power) //Determine the color of a pixel based on its state - Parameters = n, Zn{r}, Zn{i} B, P
{
	double x=log(2*log(BailoutRadius))-log(log(sqrt(fabs(ValueAtEndReal*ValueAtEndReal+ValueAtEndImaginary*ValueAtEndImaginary))))/log(Power);
	double State=(double)IterationsToEscape+x;
	return (DWORD)((15-State)/15*0xFF);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) //Process windows events
{
	switch(msg)
	{
		case WM_PAINT:
		{
			//Draw the backbuffer to the main screen
			HDC TheDC=GetDC(hWnd);
			BitBlt(TheDC,0,0,Width,Height,BackBuffer,0,0,SRCCOPY);
			ReleaseDC(hWnd, TheDC);
			break;
		}
		case WM_SIZE:
		{
			//Resize the backbuffer and redraw the window
			Width=LOWORD(lParam);
			Height=HIWORD(lParam);
			CreateBackBuffer(hWnd);
			DrawWindow(hWnd);
			break;
		}
		case WM_LBUTTONDOWN:
		{
			//Zoom in by 200% around point
			double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window scale

			//Determine the clicked point in the window
			double PointX=(double)LOWORD(lParam)/Width*ScaleX+XMin;
			double PointY=YMax-(double)HIWORD(lParam)/Height*ScaleY;

			//Determine the new scale points
			XMin=PointX-ScaleX/8;
			XMax=PointX+ScaleX/8;
			YMin=PointY-ScaleY/8;
			YMax=PointY+ScaleY/8;

			DrawWindow(hWnd); //Redraw the window
			break;
		}
		case WM_RBUTTONDOWN:
		{
			//Zoom out by 200% around point
			double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window scale

			//Determine the center of the window
			double PointX=.5*ScaleX+XMin;
			double PointY=YMax-.5*ScaleY;

			//Determine the new scale points
			XMin=PointX-ScaleX*2;
			XMax=PointX+ScaleX*2;
			YMin=PointY-ScaleY*2;
			YMax=PointY+ScaleY*2;

			DrawWindow(hWnd); //Redraw the window
			break;
		}
		case WM_CHAR:
			if(wParam=='=')
				MaxIterations*=2;
			else if(wParam=='-')
				MaxIterations/=2;
			else //Assumes a direction key was pressed
			{
				//Move in given direction
				double ScaleX=XMax-XMin, ScaleY=YMax-YMin; //Determine the current window scale

				//Determine the center of the window
				double PointX=.5*ScaleX+XMin;
				double PointY=YMax-.5*ScaleY;

				if(wParam=='w')
					PointY+=ScaleY/16;
				else if(wParam=='a')
					PointX+=ScaleX/-16;
				else if(wParam=='s')
					PointY+=ScaleY/-16;
				else if(wParam=='d')
					PointX+=ScaleX/16;

				//Determine the new scale points
				XMin=PointX-ScaleX/2;
				XMax=PointX+ScaleX/2;
				YMin=PointY-ScaleY/2;
				YMax=PointY+ScaleY/2;
			}

			DrawWindow(hWnd);
			break;
		case WM_DESTROY:
			//Terminate the program
            PostQuitMessage(0);
            return 0;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}